import operator
import math
import random
from Bullet import Bullet
class AI(object):
    '''Class for a generic enemy.'''
    def __init__(self):
        pass
    
    def run(self, enemy, game):
        pass
Stationary = AI

class Shoots(AI):
    def __init__(self, period=10, strength=None):
        super(Shoots, self).__init__()
        self.strength = strength
        self.n = 0
        self.period = period

    def run(self, enemy, game):
        self.n += 1
        if self.n % self.period == 0:
            game.shoot(enemy, Bullet(enemy, killstr=enemy.killstr, bounces=0,
                                     strength=self.strength,
                                     size=enemy.bulletsize))

class Targets(Shoots):
    def __init__(self, turnrate = 1, **kwargs):
        self.turnrate = turnrate
        super(Targets, self).__init__(**kwargs)

    def _target(self, enemy, game):
        p = game.geometry.get_pos(game.player)
        return p

    def point_to(self, target, enemy, game):
        p0 = game.geometry.get_pos(enemy)
        d = map(operator.sub, target, p0)
        x, y = d
        angle = -math.degrees(math.atan2(y, x))
        dtheta = (angle - enemy.angle)%(360)
        if abs(dtheta) < self.turnrate:
            enemy.angle = angle
        else:
            enemy.angle = (enemy.angle+cmp(180, dtheta)*self.turnrate)%360

    def run(self, enemy, game):
        self.point_to(self._target(enemy, game), enemy, game)
        super(Targets, self).run(enemy, game)

class Leads(Targets):
    def __init__(self, **kwargs):
        super(Leads, self).__init__(**kwargs)

    def _target(self, enemy, game):
        p = game.geometry.get_pos(game.player)
        p0 = game.geometry.get_pos(enemy)
        d = map(operator.sub, p, p0)
        dist = math.sqrt(d[0]*d[0]+d[1]*d[1])
        bulletspeed = 10
        if enemy.bulletspeed: bulletspeed = enemy.bulletspeed
        t = dist/bulletspeed
        v = game.geometry.get_velo(game.player)
        a = game.forces.get(game.player, (0, 0))
        # p + v*t + 1/2*a*t^2
        t0 = p[0] + v[0]*t + a[0]*t*t/2
        t1 = p[1] + v[1]*t + a[1]*t*t/2
        return [t0, t1]


class Damps(AI):
    def __init__(self, speed=1, accel=1.0/24, **kwargs):
        super(Damps, self).__init__(**kwargs)
        self.speed = speed
        self.accel = accel

    def run(self, enemy, game):
        v = game.geometry.get_velo(enemy)
        speed = math.hypot(*v)
        if speed > self.speed:
            # apply braking force of self.accel in opposite direction
            m = game.geometry.get_mass(enemy)
            diff = min(speed - self.speed, self.accel)
            brake = [-v[0]*m/speed*diff, -v[1]*m/speed*diff]
            game.add_force(enemy, brake)

class Wanders(Damps):
    CHANGING=60 # time to change directions (frames)
    def __init__(self, dur=300, **kwargs):
        super(Wanders, self).__init__(**kwargs)
        self.dur = dur

        self.states = ['travelling', 'changing direction']

        self.thisdir = (0, 0)
        self.thislong = self.dur

        self.newv = None

    def run(self, enemy, game):
        super(Wanders, self).run(enemy, game)
        if self.thislong == self.dur:
            if not self.newv:
                # pick a new direction and start applying force
                m = game.geometry.get_mass(enemy)
                dir = random.uniform(0, 360)
                dir = math.radians(-dir) # convert to "math"
                self.newv = [self.speed*math.cos(dir), self.speed*math.sin(dir)]
                #print 'self.newdir is', self.newv
                self.force = [(self.newv[0] - self.thisdir[0])*m/self.CHANGING,
                              (self.newv[1] - self.thisdir[1])*m/self.CHANGING]
                self.changing = 0
            elif self.changing == self.CHANGING:
                self.thisdir = self.newv
                self.newv = None
                self.dur = 0
            else:
                game.add_force(enemy, self.force)
                self.changing += 1

        else:
            self.dur += 1

class DoesAll(AI):
    def __init__(self, *args):
        self.ais = args

    def run(self, enemy, game):
        for ai in self.ais:
            ai.run(enemy, game)
